import numpy as np


class Sjyhwrd:

    def __init__(self, lim_age_groups, parameters, length_init):
        self.lim_age_groups = lim_age_groups
        self.nb_age_groups = len(self.lim_age_groups)
        self.parameters = parameters
        self.length_segments = length_init

    def update_state(self, state, contact_factor):
        s, j, y, h, w, r, d = self.S(state), self.J(state), self.Y(state), self.H(state), self.W(state), self.R(state), self.D(state)

        # Population updates calculus
        if j.shape[1] < y.shape[1]:
            contagious = np.hstack((np.multiply(self.parameters['zeta_j'], j),
                                    np.zeros((j.shape[0], y.shape[1] - j.shape[1])))) + \
                         np.multiply(self.parameters['zeta_y'], y)
        else:
            contagious = np.hstack((np.multiply(self.parameters['zeta_y'], y),
                                    np.zeros((y.shape[0], j.shape[1] - y.shape[1])))) + \
                         np.multiply(self.parameters['zeta_j'], j)
        densite_contag = np.sum(
            np.multiply(contact_factor, np.transpose(contagious)))

        if densite_contag == 0:
            # If there is no contagious people anymore
            force_infection = 0
        else:
            force_infection = densite_contag / (
                        np.sum(s) / np.multiply(contact_factor, self.parameters['R_0']) + densite_contag)

        # New infected, either in Y or in J
        si = np.multiply(s, force_infection)
        iy = np.multiply(si, self.parameters['theta'])
        ij = si - iy

        # Evolution of compartiment Y
        y_out = np.multiply(1 - self.parameters['eta'], y)
        yy = y - y_out
        yw = np.multiply(self.parameters['psi'].reshape(-1, 1), y_out)
        yh = y_out - yw

        # Evolution of compartiment J
        jj = j
        jr = j[:, -1]

        # Evolution of compartiment H
        h_out = np.multiply(h, 1 - self.parameters['rho'])
        hh = h - h_out
        hd = np.multiply(np.sum(h_out, axis=1), self.parameters['mu'])
        hr = np.sum(h_out, axis=1) - hd

        # Evolution of compartiment W
        w_out = np.multiply(w, 1 - self.parameters['v'])
        ww = w - w_out
        wd = np.sum(w_out, axis=1)

        # Population updates
        s_n = s - si

        y_n = np.roll(yy, shift=1, axis=1)
        y_n[:, 0] = iy

        j_n = np.roll(jj, shift=1, axis=1)
        j_n[:, 0] = ij

        h_n = np.roll(hh, shift=1, axis=1)
        h_n[:, 0] = np.sum(yh, axis=1)

        w_n = np.roll(ww, shift=1, axis=1)
        w_n[:, 0] = np.sum(yw, axis=1)

        r_n = r + jr + hr
        d_n = d + hd + wd

        return np.stack([s_n, *np.transpose(j_n), *np.transpose(y_n), *np.transpose(h_n), *np.transpose(w_n), r_n, d_n],
                        axis=0)

    def run_simulation(self, scenario, initial_conditions_category):
        sol_tot = np.zeros((scenario['t_end'] + 1,
                            self.length_segments['length_g'] +
                            self.length_segments['length_h'] +
                            self.length_segments['length_r'] +
                            self.length_segments['length_u'] + 3,
                            self.nb_age_groups))

        # Initial conditions
        sol_tot[0] = np.vstack([initial_conditions_category['S_0'],
                                np.transpose(initial_conditions_category['J_0']),
                                np.transpose(initial_conditions_category['Y_0']),
                                np.transpose(initial_conditions_category['H_0']),
                                np.transpose(initial_conditions_category['W_0']),
                                initial_conditions_category['R_0'],
                                initial_conditions_category['D_0']])

        t = 0
        political_measure_dates = np.array(scenario['political_measures']['dates'])
        while t < scenario['t_end']:
            print("Time %d" % (t + 1))
            # Get right contact factor
            idx = np.sum(t >= political_measure_dates)
            contact_factor = scenario['political_measures']['contact_factor'][idx-1]
            sol_tot[t + 1] = self.update_state(sol_tot[t], contact_factor)
            t += 1

        print('Simulation complete.')
        return sol_tot

    def get_N_0(self):
        return self.initial_conditions['N_0']

    ######################
    ## Compartments from state
    ######################

    def get_state_order(self):
        return ['S',
                'J',
                *['J_'] * (self.length_segments['length_g'] - 1),
                'Y',
                *['Y_'] * (self.length_segments['length_h'] - 1),
                'H',
                *['H_'] * (self.length_segments['length_r'] - 1),
                'W',
                *['W_'] * (self.length_segments['length_u'] - 1),
                'R',
                'D']

    def S(self, state):
        return state[self.get_state_order().index("S")]

    def J(self, state):
        idx = self.get_state_order().index("J")
        return np.transpose(state[idx: idx + self.length_segments['length_g']])

    def Y(self, state):
        idx = self.get_state_order().index("Y")
        return np.transpose(state[idx: idx + self.length_segments['length_h']])

    def H(self, state):
        idx = self.get_state_order().index("H")
        return np.transpose(state[idx: idx + self.length_segments['length_r']])

    def W(self, state):
        idx = self.get_state_order().index("W")
        return np.transpose(state[idx: idx + self.length_segments['length_u']])

    def R(self, state):
        return state[self.get_state_order().index("R")]

    def D(self, state):
        return state[self.get_state_order().index("D")]

    def N(self, state):
        return self.S(state) + \
               np.sum(self.J(state) + self.Y(state) + self.H(state) + self.W(state), axis=1) + \
               self.R(state)
















